/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

#ifndef DRVHY8415IPASYN_H
#define DRVHY8415IPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the DAC memory layout. */
#define BANKSIZE  				0x100000       	/* maximum memory 1M x 16bits */
#define CHANNELBANKSIZE 		0x8000       	/* maximum memory per channel */
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_DAC_CHANNELS    	16      		/* number of DAC channels on 8415 card */
#define PI						3.1419			/* pi value */
#define MAX_CARDS          		256          /* maxiumum number of cards in a system */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4"
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0x8415

/* define bits in the 8415 CSR */
#define CSR_HF 					0x0001          /* STAUTS: indicate finishes of half full updates */
#define CSR_F	 				0x0002          /* STAUTS: indicate finishs of full updates */
#define CSR_CC	 				0x0004          /* STAUTS: indicate conversions complete */
#define CSR_CONT				0x0008          /* Set continuous function generation */
#define CSR_MH 					0x0010          /* STATUS: "1" indicates that memory is inhabited from Strobe line */
#define CSR_1MB 				0x0020          /* using 1MB memory when set or using 2MB when not set */
#define CSR_SSU 				0x0040          /* Writing "1" sets all DAC outputs to zero */
#define CSR_BS    				0x0080          /* STATUS: Flag showing unit is busy shifting data to DAC */
#define CSR_HE     				0x0100          /* Enable interrupt on lower memory has been filled */
#define CSR_FE     				0x0200          /* Enable interrupt on upper memory has been filled */
#define CSR_EE     				0x0400          /* Enable interrupt at the end of number of updates */
#define CSR_ET      			0x0800          /* Enable Strobe line when set to stop updating DAC from memory */
#define CSR_XC      			0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      			0x2000          /* Software trigger. To trigger, do "clear" then "set" */
#define CSR_EX      			0x4000          /* 0 => register mode, 1 => trigger and memory mode. */
#define CSR_ARM     			0x8000          /* ARMing causes the values loaded in to the DAC. cleared afterwards */

/* define extension csr */
#define ACR_PM10V	 			0x0000          /* Range setting: +/-10V */
#define ACR_PM5V	 			0x0001          /* Range setting: +/-5V */
#define ACR_ZT10V	 			0x0002          /* Range setting: 0 -10V */
#define ACR_ZT5V	 			0x0003          /* Range setting: 0 -5V */
#define ACR_BTC	 				0x0004          /* 0 = straight binary, 1 = 2's complement */
#define ACR_MT	 				0x0010          /* Multi trigger mode */
#define ACR_RP					0x0020          /* Repeat pattern in multi trigger mode */
#define ACR_CAL					0x4000          /* only for register mode, set 1 means do not use on board calibration */
#define ACR_EN16				0x8000          /* set 1 to use 16bit DAC, 0 (default) to use 18bit DAC */

#define ENABLE 					1
#define DISABLE 				0

#define INIT 					0
#define RUNTIME 				1

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0003   		/* support register and trigger modes */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_UPDATING		1           	/* triggerred and collecting data */
#define UPDATE_COMPLETE     	2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{
    USHORT csr;                  	/* 0x00 */
    USHORT pointer;                	/* 0x02 */
    USHORT updateNo;	     		/* 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT extcsr;           		/* 0x0A */
    USHORT dpdr;	               	/* 0x0C, internal use */
    USHORT notused1;             	/* 0x0E */
    USHORT data[MAX_DAC_CHANNELS*2];/* 0x10 - 0x4E, for 16bit mode, it only uses 0x10-0x2E */
    USHORT notused2[8];				/* 0x50 - 0x5E */
    USHORT dpcr[4];            		/* 0x60 - 0x66, digital pot calibration factors, internal use */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */
    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
    USHORT CAL[0xBC - 0x9E];		/* store calibration data */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8415Register,
	hy8415Trigger,
} hy8415Modes;

/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int cardNumber ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
}

/** This is the class definition for the Hy8415Asyn class
  */
class Hy8415Asyn : public asynPortDriver {
public:
	Hy8415Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode, 
			int range, int clockRate, int extClock);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus writeInt32Array(asynUser *pasynUser, epicsInt32 *value,
                                        size_t nElements);
    virtual asynStatus readInt32Array(asynUser *pasynUser, epicsInt32 *value,
                                        size_t nElements, size_t *nIn);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
    virtual asynStatus writeFloat64(asynUser *pasynUser, epicsFloat64 value);
    virtual asynStatus writeFloat64Array(asynUser *pasynUser, epicsFloat64 *value,
                                        size_t nElements);
    virtual void report(FILE *fp, int details);
    
    int Hy8415ExtInit(int continuous, int format, int multitrigger, 
    		int repeatpattern, int set16bit);
    int Hy8415Function(int type, double volt, int samples, int mask);
    void rebootCallback();
	void intSR(); 				/* Interrupt Service Routine */
	void intQueuedTask();

protected:
    int hy8415Data_;
    #define FIRST_8415_PARAM hy8415Data_
	int hy8415FloatData_;
	int hy8415SetMode_;
	int hy8415NoOfUpdate_;
	int hy8415Continuous_;
	int hy8415SoftTrigger_;
	int hy8415SetClockRate_;
	int hy8415SetExtClock_;
	int hy8415Range_;
	int hy8415SetFormat_;
	int hy8415SetMultiTrigger_;
	int hy8415RepeatPattern_;
	int hy8415Set16bit_;
	int hy8415ZeroDAC_;

	int hy8415Counter_;
	int hy8415TriggerStatus_;
	int hy8415FWVersion_;
	int hy8415DriverVersion_;
	int hy8415Support_;
    #define LAST_8415_PARAM hy8415Support_

private:
	void setupDAC               		();
	asynStatus setClockRate            	( int rate );
	asynStatus setExtClock             	( int extClock );
	asynStatus softwareTrigger         	();
	asynStatus setMode           		( int mode );
	asynStatus setRange           		( int range );
	asynStatus setNumberUpdate         	( int number );
	asynStatus setContinuousUpdate     	( int cont );
	asynStatus setFormat          		( int format );
	asynStatus setMultiTrigger    		( int multi );
	asynStatus setRepeat          		( int repeat );
	asynStatus set16Bit          		( int bit );
	int fromVoltage             		( double value );
	asynStatus zeroDAC          		();
	asynStatus resetTriggerState		();
    
	volatile IO_MEMORY 	*dac;			   	/* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT 	*convMemoryBase;   	/* pointer to IP card memory space */

	/* Storage for the current Configuration parameters */
    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum;		                /* Interrupt vector number - boot time configuration only */
    USHORT	mode;		                	/* mode, 0 for continuous mode, 1 for trigger mode */
    USHORT	range;		                	/* range setting, 0 for -10v ~ +10V, 1 for -5V ~ +5V */
    USHORT	format;		                	/* data format, 0 for 2's complement, 1 for straight */
    USHORT	clockrate;	                	/* clock rate */
    USHORT	extclock;		               	/* 0 internal clock, 1 for external clock */
    int 	noofupdate;
    int		continuous;
    int 	multitrigger;
    int		repeatpattern;
    int		bitFormat;

/* driver specific parameters */
    int		rebooting;		                /* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
};

#define hy8415DataString	        "DATA"
#define hy8415FloatDataString	    "FLOATDATA"
#define hy8415SetModeString		    "SETMODE"
#define hy8415NoOfUpdateString	    "NOOFUPDATE"
#define hy8415ContinuousString	    "CONTINUOUS"
#define hy8415SoftTriggerString		"SOFTTRIGGER"
#define hy8415SetClockRateString	"SETCLOCKRATE"
#define hy8415SetExtClockString		"SETEXTCLOCK"
#define hy8415RangeString	    	"RANGE"
#define hy8415SetFormatString	    "FORMAT"
#define hy8415SetMultiTriggerString	"MULTITGR"
#define hy8415RepeatPatternString	"REPEATPTN"
#define hy8415Set16bitString		"SET16BIT"
#define hy8415ZeroDACString			"ZERODAC"

#define hy8415CounterString			"COUNTER"
#define hy8415TriggerStatusString	"TRIGGERSTATE"
#define hy8415FWVersionString		"FWVERSION"
#define hy8415DriverVersionString	"DRIVERVERSION"
#define hy8415SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8415_PARAM - &FIRST_8415_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *         mode = 0: register, 1: triggered
 *      range   = 0: -10V ~ +10V, 1:-5V ~ +5V, 2: 0 ~ 10V, 3: 0 ~ 5V 
 *    clockRate = 0: 1Hz,    1: 2Hz,    2: 5Hz,    3: 10Hz,
 *				  4: 20Hz,   5: 50Hz,   6: 100Hz,  7: 200Hz,
 *				  8: 500Hz,  9: 1kHz,  10: 2kHz,  11: 5kHz,
 *				  12: 10khz, 13: 20Khz, 14: 50kHz
 *     extClock = 0: internal, 1: extrenal,
 */
extern "C" int Hy8415AsynInit(char *portName, int vmeSlotNum, int ipSlotNum, int vectorNum, 
					int mode, int range, int clockRate, int extClock)
{
    Hy8415Asyn *pHy8415Asyn = new Hy8415Asyn(portName, vmeSlotNum, ipSlotNum, vectorNum, mode, 
    		range, clockRate, extClock);
    pHy8415Asyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

/** Extension Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   noofupdate = number of update from the memory 
 *   continuous = 1: continuous update from memory, 0: update once
 *    	 format = 0: straight binary, 1: 2's complement
 * multitrigger = 0: normal mode, 1: multi trigger mode
 * epeatpattern = 1: repeat pattern in multi trigger mode, 0: do not repeat 
 *	   set16bit = 0: 18bit DAC, 1: 16bit DAC
 */
extern "C" int Hy8415AsynExtInit(char *portName, int continuous, int format, 
		int multitrigger, int repeatpattern, int set16bit)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8415Asyn *pHy8415 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8415AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    
    //find AN interface in order to derive the drvPvt
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8415AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8415 = (Hy8415Asyn*) pasynInterface->drvPvt;
    pHy8415->Hy8415ExtInit(continuous, format, multitrigger, repeatpattern, set16bit);
    return(asynSuccess);
} 

/** Function genertor configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *         type = 0: sin wave, 1: pulse train
 *         volt = peak value of voltage 
 *      samples = samples per cycle
 *  channelmask = channel mask. 0: no function, 1: yes
 */
extern "C" int Hy8415Function(char *portName, int type, double volt, int samples, int channelmask)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8415Asyn *pHy8415 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8402AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8402AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8415 = (Hy8415Asyn*) pasynInterface->drvPvt;
    pHy8415->Hy8415Function(type, volt, samples, channelmask);
    return(asynSuccess);
}
				
// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "vectorNum",iocshArgInt};
static const iocshArg initArg4 = { "mode", iocshArgInt};
static const iocshArg initArg5 = { "range", iocshArgInt};
static const iocshArg initArg6 = { "clockRate", iocshArgInt};
static const iocshArg initArg7 = { "extClock", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4,
                                             &initArg5,
                                             &initArg6,
                                             &initArg7};
static const iocshFuncDef initFuncDef = {"Hy8415AsynInit",8,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8415AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival, 
					args[7].ival);
}

// Extension init
static const iocshArg initArg0Ext = { "Port name",iocshArgString};
static const iocshArg initArg1Ext = { "continuous",iocshArgInt};
static const iocshArg initArg2Ext = { "format",iocshArgInt};
static const iocshArg initArg3Ext = { "multi Trigger",iocshArgInt};
static const iocshArg initArg4Ext = { "repeat Patern",iocshArgInt};
static const iocshArg initArg5Ext = { "Set 16bit", iocshArgInt};
static const iocshArg * const initArgsExt[] = {&initArg0Ext,
                                             &initArg1Ext,
                                             &initArg2Ext,
                                             &initArg3Ext,
                                             &initArg4Ext,
                                             &initArg5Ext};
static const iocshFuncDef initFuncDefExt = {"Hy8415AsynExtInit",6,initArgsExt};
static void initCallFuncExt(const iocshArgBuf *args)
{
    Hy8415AsynExtInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival);
}

// Function Generator
static const iocshArg funcArg0 = { "Port name",iocshArgString};
static const iocshArg funcArg1 = { "type",iocshArgInt};
static const iocshArg funcArg2 = { "volt",iocshArgDouble};
static const iocshArg funcArg3 = { "samples",iocshArgInt};
static const iocshArg funcArg4 = { "cahnnelMask",iocshArgInt};
static const iocshArg * const funcArgs[] = {&funcArg0,
                                             &funcArg1,
                                             &funcArg2,
                                             &funcArg3,
                                             &funcArg4};
static const iocshFuncDef funcFuncDef = {"Hy8415Function",5,funcArgs};
static void funcCallFunc(const iocshArgBuf *args)
{
    Hy8415Function(args[0].sval, args[1].ival, args[2].dval, args[3].ival, args[4].ival);
}

void drvHy8415AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
    iocshRegister(&initFuncDefExt,initCallFuncExt);
    iocshRegister(&funcFuncDef,funcCallFunc);
}

extern "C" {
epicsExportRegistrar(drvHy8415AsynRegister);
}

